import csv
from datetime import datetime

import typer

from .cache import cache
from .geocode import Geocoders, geocode_line, geocode_proxy
from .io import Formats, get_reader, get_writer
from .model import GeocodingResult, get_address

cli = typer.Typer()
cli_cache = typer.Typer()
cli.add_typer(cli_cache, name="cache")


@cli.command()
def format_line(
    input_file: typer.FileText = typer.Option("-", "-i", help="input file"),
    output_file: typer.FileTextWrite = typer.Option("-", "-o", help="output file"),
    header: bool = typer.Option(True, help="Input stream has csv header row"),
):
    reader = get_reader(input_file, Formats.csv, header=header)
    writer = csv.writer(output_file)

    for address, country, language, *rest in reader:
        address = get_address(address, language=language, country=country)
        writer.writerow(
            [address.get_formatted_line(), ";".join(address.country), *rest]
        )


@cli.command()
def geocode(
    input_file: typer.FileText = typer.Option("-", "-i", help="Input file"),
    input_format: Formats = typer.Option(Formats.ftm.value, help="Input format"),
    output_file: typer.FileTextWrite = typer.Option("-", "-o", help="Output file"),
    output_format: Formats = typer.Option(Formats.ftm.value, help="Output format"),
    geocoder: list[Geocoders] = typer.Option(
        [Geocoders.nominatim.value], "--geocoder", "-g"
    ),
    cache: bool = typer.Option(True, help="Use cache database"),
    include_raw: bool = typer.Option(
        False, help="Include geocoder raw response (for csv output only)"
    ),
    rewrite_ids: bool = typer.Option(
        True, help="Rewrite `Address` entity ids to canonized id"
    ),
    header: bool = typer.Option(True, help="Input stream has csv header row"),
):
    reader = get_reader(input_file, input_format, header=header)
    writer = get_writer(output_file, output_format, include_raw=include_raw)

    if input_format == Formats.ftm:
        for proxy in reader:
            for result in geocode_proxy(
                geocoder,
                proxy,
                use_cache=cache,
                output_format=output_format,
                rewrite_ids=rewrite_ids,
            ):
                writer(result)

    else:
        for address, country, language, *rest in reader:
            result = geocode_line(geocoder, address, use_cache=cache, country=country)
            if result is not None:
                writer(result, *rest)


@cli_cache.command("iterate")
def cache_iterate(
    output_file: typer.FileTextWrite = typer.Option("-", "-o", help="Output file"),
    output_format: Formats = typer.Option(Formats.ftm.value, help="Output format"),
    include_raw: bool = typer.Option(False, help="Include geocoder raw response"),
):
    writer = get_writer(output_file, output_format, include_raw=include_raw)

    for res in cache.iterate():
        writer(res)


@cli_cache.command("populate")
def cache_populate(
    input_file: typer.FileText = typer.Option("-", "-i", help="Input file"),
    input_format: Formats = typer.Option(Formats.csv.value, help="Input format"),
):
    reader = csv.DictReader(input_file)
    bulk = cache.bulk()

    for row in reader:
        if "ts" not in row:
            row["ts"] = datetime.now()
        result = GeocodingResult(**row)
        bulk.put(result)
    bulk.flush()



import csv
from enum import Enum
from typing import Generator, Literal

import orjson
import typer
from followthemoney import model
from followthemoney.proxy import EntityProxy

from .model import Address, GeocodingResult


class Formats(Enum):
    csv = "csv"
    ftm = "ftm"


Row = tuple[str, str | None, str | None, ...]


def dump_proxy(proxy: EntityProxy) -> str:
    return orjson.dumps(proxy.to_dict(), option=orjson.OPT_APPEND_NEWLINE).decode()


def read_ftm(input_file: typer.FileText) -> Generator[Row, None, None]:
    for row in input_file.readlines():
        yield model.get_proxy(orjson.loads(row))


def write_ftm(output_file: typer.FileTextWrite):
    def _write(data: GeocodingResult | EntityProxy):
        if isinstance(data, GeocodingResult):
            address = Address.from_result(data)
            proxy = address.to_proxy()
            output_file.write(dump_proxy(proxy))
        else:
            output_file.write(dump_proxy(data))

    return _write


def read_csv(
    input_file: typer.FileText, header: bool = True
) -> Generator[Row, None, None]:
    reader = csv.reader(input_file)
    if header:
        next(reader)
    for row in reader:
        address, *rest = row
        country, language = None, None
        if len(rest):
            country, *rest = rest
            if len(rest):
                language, *rest = rest
            else:
                language = country
        yield address, country, language, *rest


def write_csv(output_file: typer.FileTextWrite, include_raw: bool | None = False):
    fieldnames = GeocodingResult.__annotations__.keys()
    if not include_raw:
        fieldnames = [
            f for f in fieldnames if f not in ("ts", "geocoder_raw", "cache_key")
        ]

    writer = csv.DictWriter(output_file, fieldnames=fieldnames)
    writer.writeheader()

    def _write(result: GeocodingResult, *rest):
        result = {k: v for k, v in result if k in fieldnames}
        writer.writerow(result)

    return _write


def get_reader(
    input_file: typer.FileText, input_format: Formats, **kwargs
) -> Literal[read_ftm, read_csv]:
    if input_format == Formats.ftm:
        return read_ftm(input_file)
    if input_format == Formats.csv:
        return read_csv(input_file, **kwargs)


def get_writer(
    output_file: typer.FileTextWrite, output_format: Formats, **kwargs
) -> Literal[write_ftm, write_csv]:
    if output_format == Formats.ftm:
        return write_ftm(output_file)
    if output_format == Formats.csv:
        return write_csv(output_file, **kwargs)



import logging
import os

GEOCODERS = tuple(os.environ.get("GEOCODERS", "nominatim").split(","))
CACHE_TABLE = os.environ.get("FTMGEO_CACHE_TABLE", "ftmgeo_cache")

USER_AGENT = os.environ.get("FTMGEO_USER_AGENT", "ftm-geocode")
DEFAULT_TIMEOUT = os.environ.get("FTMGEO_DEFAULT_TIMEOUT", 10)
MIN_DELAY_SECONDS = float(os.environ.get("FTMGEO_MIN_DELAY_SECONDS", 0.1))
MAX_RETRIES = int(os.environ.get("FTMGEO_MAX_RETRIES", 5))
LOG_LEVEL = logging.getLevelName(os.environ.get("LOG_LEVEL", "info").upper())
LOG_JSON = False



from functools import cache
from typing import Generator

import dataset
from dataset.database import Database
from dataset.table import Table
from followthemoney.util import make_entity_id
from ftmstore.settings import DATABASE_URI
from normality import normalize

from .logging import get_logger
from .model import AddressInput, GeocodingResult, PostalContext, get_address_id
from .settings import CACHE_TABLE

log = get_logger(__name__)


def get_cache_key(value: str, **ctx: PostalContext) -> str:
    country = ctx.get("country")
    ident = make_entity_id(normalize(value))
    if country is not None:
        return f"{country.lower()}-{ident}"
    return ident


@cache
def get_connection() -> Database:
    return dataset.connect(DATABASE_URI)


class BulkWrite:
    limit = 10_000
    rows: list[GeocodingResult] = []
    i = 0

    def put(self, row: GeocodingResult):
        row.cache_key = get_cache_key(row.original_line)
        self.rows.append(dict(row))
        self.i += 1
        if self.i % self.limit == 0:
            self.flush()

    def flush(self):
        with get_connection() as tx:
            tx[CACHE_TABLE].insert_many(self.rows)
            self.rows = []
            log.info(f"Cache: insert {self.i} results.")


class Cache:
    def __str__(self) -> str:
        return f"<GeoCache `{DATABASE_URI}`>"

    def get_table(self) -> Table:
        db = get_connection()
        return db[CACHE_TABLE]

    def put(self, row: GeocodingResult):
        bulk = self.bulk()
        bulk.put(row)
        bulk.flush()

    def bulk(self) -> BulkWrite:
        return BulkWrite()

    def get(self, data: AddressInput, **ctx) -> GeocodingResult | None:
        table = self.get_table()
        if isinstance(data, str):
            for res in table.find(cache_key=get_cache_key(data, **ctx)):
                return GeocodingResult(**res)
        address_id = get_address_id(data, **ctx)
        for res in table.find(address_id=address_id):
            return GeocodingResult(**res)
        for res in table.find(canonical_id=address_id):
            return GeocodingResult(**res)

    def iterate(self) -> Generator[GeocodingResult, None, None]:
        with get_connection() as tx:
            for row in tx[CACHE_TABLE]:
                yield GeocodingResult(**row)


def get_cache() -> Cache:
    return Cache()


cache = get_cache()



import os
from datetime import datetime
from enum import Enum
from typing import Any, Generator, TypedDict

import geopy.geocoders
import orjson
from banal import clean_dict
from followthemoney import model
from followthemoney.proxy import EntityProxy
from geopy.adapters import AdapterHTTPError
from geopy.exc import GeocoderQueryError
from geopy.extra.rate_limiter import RateLimiter
from geopy.geocoders import SERVICE_TO_GEOCODER, get_geocoder_for_service

from . import settings
from .cache import cache
from .io import Formats
from .logging import get_logger
from .model import Address, GeocodingResult, get_address_id
from .util import apply_address, get_proxy_addresses, normalize, normalize_google

geopy.geocoders.options.default_user_agent = settings.USER_AGENT
geopy.geocoders.options.default_timeout = settings.DEFAULT_TIMEOUT

log = get_logger(__name__)

Geocoders = Enum("Geocoders", ((k, k) for k in SERVICE_TO_GEOCODER.keys()))


class GeocodingContext(TypedDict):
    country: str | None = None


CONFIG = {
    Geocoders.nominatim: {
        "domain": os.environ.get("FTMGEO_NOMINATIM_DOMAIN"),
    },
    Geocoders.googlev3: {
        "api_key": os.environ.get("FTMGEO_GOOGLE_API_KEY"),
    },
}

GEOCODER_CTX = {
    Geocoders.nominatim: lambda ctx: {"country_codes": ctx.get("country")},
    Geocoders.googlev3: lambda ctx: {"region": ctx.get("country")},
    Geocoders.arcgis: lambda ctx: {"out_fields": "*"},
}


def get_geocoder(geocoder: Geocoders):
    config = clean_dict(CONFIG.get(geocoder, {}))
    geocoder = get_geocoder_for_service(geocoder.value)
    return geocoder(**config)


def _geocode(
    geocoder: Geocoders, value: str, **ctx: GeocodingContext
) -> GeocodingResult | None:
    geocoding_value = value
    if "google" in geocoder.value:
        geocoding_value = normalize_google(value)
    geocode_ctx = {}
    ctx_getter = GEOCODER_CTX.get(geocoder)
    if ctx_getter is not None:
        geocode_ctx = ctx_getter(ctx)
    geolocator = get_geocoder(geocoder)
    geocode = RateLimiter(
        geolocator.geocode,
        min_delay_seconds=settings.MIN_DELAY_SECONDS,
        max_retries=settings.MAX_RETRIES,
    )
    address_id = get_address_id(value, **ctx)

    try:
        result = geocode(geocoding_value, **geocode_ctx)
    except (AdapterHTTPError, GeocoderQueryError):
        result = None
        pass  # error will be logged

    if result is not None:
        log.info(f"Geocoder hit: `{value}`", geocoder=geocoder.value)
        address = Address.from_string(result.address, **ctx)
        result = GeocodingResult(
            address_id=address_id,
            canonical_id=address.get_id(),
            original_line=value,
            result_line=result.address,
            country=address.get_country(),
            lat=result.latitude,
            lon=result.longitude,
            geocoder=geocoder.value,
            geocoder_place_id=result.raw.get("place_id"),
            geocoder_raw=orjson.dumps(result.raw),
            ts=datetime.now(),
        )
        cache.put(result)
        return result


def geocode_line(
    geocoder: list[Geocoders],
    value: str,
    use_cache: bool | None = True,
    **ctx: GeocodingContext,
) -> GeocodingResult | None:

    value = normalize(value)

    # look in cache
    if use_cache:
        result = cache.get(value)
        if result is not None:
            log.info(f"Cache hit: `{value}`", cache=str(cache))
            return result

    # geocode
    geocoders = geocoder
    for geocoder in geocoders:
        result = _geocode(geocoder, value, **ctx)
        if result is not None:
            return result

    log.warning(f"Not found: `{value}`", geocoder=geocoder)


def geocode_proxy(
    geocoder: list[Geocoders],
    proxy: EntityProxy | dict[str, Any],
    use_cache: bool | None = True,
    output_format: Formats | None = Formats.ftm,
    rewrite_ids: bool | None = True,
) -> Generator[EntityProxy | GeocodingResult, None, None]:
    proxy = model.get_proxy(proxy)
    is_address = proxy.schema.is_a("Address")
    ctx = {"country": proxy.first("country") or ""}
    results = (
        geocode_line(geocoder, value, use_cache=use_cache, **ctx)
        for value in get_proxy_addresses(proxy)
    )
    if output_format == Formats.ftm:
        for result in results:
            if result is not None:
                address = Address.from_result(result)
                address = address.to_proxy()
                proxy = apply_address(proxy, address, rewrite_id=rewrite_ids)
                if is_address:
                    yield proxy
                else:
                    yield address
        if not is_address:
            yield proxy
    else:
        for result in results:
            if result is not None:
                yield result



from collections import defaultdict
from datetime import datetime
from typing import Any, TypedDict

import lazy_import
from banal import clean_dict
from followthemoney import model
from followthemoney.proxy import E
from followthemoney.util import make_entity_id
from normality import collapse_spaces, normalize
from pydantic import BaseModel, create_model
from zavod.parse.addresses import format_line

from .util import clean_country_codes, clean_country_names, get_country_code, get_first


class GeocodingResult(BaseModel):
    cache_key: str | None = None
    address_id: str
    canonical_id: str
    original_line: str
    result_line: str
    country: str
    lat: float
    lon: float
    geocoder: str
    geocoder_place_id: str | None = None
    geocoder_raw: str | None = None
    ts: datetime


# https://github.com/openvenues/libpostal#parser-labels
# postal -> ftm
# FIXME extend ftm schema to align with postal output?
MAPPING = (
    # venue name e.g. "Brooklyn Academy of Music", and building names e.g. "Empire State Building"
    ("house", "remarks"),
    # for category queries like "restaurants", etc.
    ("category", "keywords"),
    # phrases like "in", "near", etc. used after a category phrase to help with parsing queries like "restaurants in Brooklyn"
    ("near", "remarks"),
    # usually refers to the external (street-facing) building number. In some countries this may be a compount, hyphenated number which also includes an apartment number, or a block number (a la Japan), but libpostal will just call it the house_number for simplicity.
    ("house_number", "remarks"),
    # street name(s)
    ("road", "street"),
    # an apartment, unit, office, lot, or other secondary unit designator
    ("unit", "remarks"),
    # expressions indicating a floor number e.g. "3rd Floor", "Ground Floor", etc.
    ("level", "remarks"),
    # numbered/lettered staircase
    ("staircase", "remarks"),
    # numbered/lettered entrance
    ("entrance", "remarks"),
    # post office box: typically found in non-physical (mail-only) addresses
    ("po_box", "postOfficeBox"),
    # postal codes used for mail sorting
    ("postcode", "postalCode"),
    # usually an unofficial neighborhood name like "Harlem", "South Bronx", or "Crown Heights"
    ("suburb", "remarks"),
    # these are usually boroughs or districts within a city that serve some official purpose e.g. "Brooklyn" or "Hackney" or "Bratislava IV"
    ("city_district", "remarks"),
    # any human settlement including cities, towns, villages, hamlets, localities, etc.
    ("city", "city"),
    # named islands e.g. "Maui"
    ("island", "region"),
    # usually a second-level administrative division or county.
    ("state_district", "region"),
    # a first-level administrative division. Scotland, Northern Ireland, Wales, and England in the UK are mapped to "state" as well (convention used in OSM, GeoPlanet, etc.)
    ("state", "state"),
    # informal subdivision of a country without any political status
    ("country_region", "region"),
    # sovereign nations and their dependent territories, anything with an ISO-3166 code.
    ("country", "country"),
    # currently only used for appending “West Indies” after the country name, a pattern frequently used in the English-speaking Caribbean e.g. “Jamaica, West Indies”
    ("world_region", "region"),
)

Values = list[str] | None


class PostalContext(TypedDict):
    language: str | None = ""
    country: str | None = ""


class AddressBase(BaseModel):
    def get_country(self) -> str:
        return ";".join(self.country or [])

    def get_first(self, attr, default: Any | None = None) -> str | None:
        return get_first(getattr(self, attr, None), default)

    def get_id(self) -> str:  # serves as cache key
        ident = make_entity_id(normalize(self.get_formatted_line()))
        country = self.get_first("country")
        if country:
            ident = f"{country.lower()}-{ident}"
        return f"addr-{ident}"

    def to_dict(self) -> dict[str, list[str]]:
        return clean_dict(dict(self))


class PostalAddressBase(AddressBase):
    country_code: Values = None

    def __init__(self, **data):
        data["country"] = clean_country_names(data.get("country"))
        data["country_code"] = clean_country_codes(data.get("country"))
        super().__init__(**data)

    def get_formatted_line(self) -> str:
        return format_line(
            summary=collapse_spaces(
                " ".join(
                    (
                        self.get_first("house", ""),
                        self.get_first("house_number", ""),
                        self.get_first("near", ""),
                    )
                )
            )
            or None,  # noqa
            po_box=self.get_first("po_box"),
            street=self.get_first("street"),
            postal_code=self.get_first("postcode"),
            city=self.get_first("city"),
            state=self.get_first("state"),
            country_code=self.get_first("country_code"),
        )

    @classmethod
    def from_postal_result(
        cls, input_data: list[tuple[str, str]], **ctx: PostalContext
    ) -> "PostalAddress":
        data = defaultdict(set)
        for value, key in input_data:
            data[key].add(value)
        if "country" in ctx:
            data["country"].add(ctx["country"])
        return cls(**data)

    @classmethod
    def from_string(cls, value: str, **ctx: PostalContext) -> "PostalAddress":
        parse_address = lazy_import.lazy_callable("postal.parser.parse_address")
        # postal screams if language or country is None
        ctx = {k: ctx.get(k, "") for k in ("language", "country")}
        result = parse_address(value, **ctx)
        return cls.from_postal_result(result, **ctx)


PostalAddress: PostalAddressBase = create_model(
    "PostalAddress",
    **{m[0]: (Values, None) for m in MAPPING},
    __base__=PostalAddressBase,
)

FtmAddressBase: AddressBase = create_model(
    "FtmAddress",
    **{p: (Values, None) for p in model.get("Address").properties},
    __base__=AddressBase,
)


class Address(FtmAddressBase):
    def to_proxy(self) -> E:
        return model.get_proxy(
            {
                "id": self.get_id(),
                "schema": "Address",
                "properties": clean_dict(self.dict()),
            }
        )

    def get_formatted_line(self) -> str:
        return format_line(
            summary=collapse_spaces(
                " ".join((self.get_first("summary", ""), " ".join(self.remarks or [])))
            ),
            po_box=self.get_first("postOfficeBox"),
            street=self.get_first("street"),
            postal_code=self.get_first("postalCode"),
            city=self.get_first("city"),
            state=self.get_first("state"),
            country_code=get_country_code(self.get_first("country")),
        )

    @classmethod
    def from_postal(cls, input_data: PostalAddress, **ctx: PostalContext) -> "Address":
        mapping = dict(MAPPING)
        data = defaultdict(set)
        data["country"].add(ctx.get("country"))
        for key, values in input_data:
            if key == "country_code":
                key = "country"
            if values is not None:
                data[mapping[key]].update(values)
        data["country"] = clean_country_codes(data["country"])
        return cls(**data)

    @classmethod
    def from_string(cls, value: str, **ctx: PostalContext) -> "Address":
        return cls.from_postal(PostalAddress.from_string(value, **ctx))

    @classmethod
    def from_result(cls, result: GeocodingResult) -> "Address":
        ctx = {"country": result.country}
        address = cls.from_postal(PostalAddress.from_string(result.result_line, **ctx))
        address.full = [result.result_line]
        address.latitude = [str(result.lat)]
        address.longitude = [str(result.lon)]
        return address


AddressInput = str | Address | PostalAddress


def get_address(data: AddressInput, **ctx: PostalContext) -> Address:
    if isinstance(data, str):
        return Address.from_string(data, **ctx)
    if isinstance(data, PostalAddress):
        return Address.from_postal(data, **ctx)
    return data


def get_formatted_line(data: AddressInput, **ctx: PostalContext) -> str:
    address = get_address(data, **ctx)
    return address.get_formatted_line()


def get_address_id(data: AddressInput, **ctx: PostalContext) -> str:
    address = get_address(data, **ctx)
    return address.get_id()
